/*
 * JBoss, Home of Professional Open Source.
 * Copyright 2008, Red Hat Middleware LLC, and individual contributors
 * as indicated by the @author tags. See the copyright.txt file in the
 * distribution for a full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

package org.jboss.cache.integration.websession;

import java.util.Collections;

import org.jboss.cache.commands.write.PutDataMapCommand;
import org.jboss.cache.commands.write.RemoveNodeCommand;
import org.jboss.cache.integration.websession.util.BuddyReplicationAssertions;
import org.jboss.cache.integration.websession.util.GetAttributesServlet;
import org.jboss.cache.integration.websession.util.InvalidationServlet;
import org.jboss.cache.integration.websession.util.MultipleActionServlet;
import org.jboss.cache.integration.websession.util.Request;
import org.jboss.cache.integration.websession.util.SessionManager;
import org.jboss.cache.integration.websession.util.SetAttributesServlet;
import org.jboss.cache.integration.websession.util.WebSessionTestBase;
import org.jboss.cache.util.internals.replicationlisteners.ReplicationListener;
import org.testng.annotations.Test;


/**
 * @author Brian Stansberry
 *
 */
@Test(groups = "integration", sequential = true, testName = "integration.websession.BuddyReplicationFailoverTest")
public class BuddyReplicationFailoverTest extends WebSessionTestBase
{
   public static final String KEY = "key";

   @Override
   protected String getCacheConfigName()
   {
      return "br-standard-session-cache";
   }

   @Override
   protected int getNumCacheManagers()
   {
      return 4;
   }
   
   @Override
   protected boolean getCreateManagersInSetup()
   {
      return true;
   }
   
   public void testFailoverAndImmediateInvalidate()
   {
      int attr = 0;
      
      SessionManager mgr0 = getSessionManagers().get(0);
      SessionManager mgr1 = getSessionManagers().get(1);
      SessionManager mgr2 = getSessionManagers().get(2);
      SessionManager mgr3 = getSessionManagers().get(3);
      
      String contextHostName = mgr0.getContextHostName();

      // Create the session
      SetAttributesServlet sas = new SetAttributesServlet(Collections.singletonMap(KEY, getAttributeValue(attr++)));
      ReplicationListener replListener1 = getReplicationListener(mgr1.getCache());
      replListener1.expectWithTx(PutDataMapCommand.class);
      
      Request req = new Request(mgr0, null, sas);
      req.execute();      
      replListener1.waitForReplicationToOccur();      
      
      String sessionId = sas.getSessionId();
      assert sessionId != null : "session id is null";
      // validate cache contents
      BuddyReplicationAssertions.assertBuddyBackup(contextHostName, sessionId, mgr0.getCache(), mgr1.getCache());
      
      // Modify the session
      sas = new SetAttributesServlet(Collections.singletonMap(KEY, getAttributeValue(attr++))); 
      replListener1.expectWithTx(PutDataMapCommand.class);
      
      req = new Request(mgr0, sessionId, sas);
      req.execute();      
      replListener1.waitForReplicationToOccur();
      
      // Fail over; request reads the session and then modifies the session
      GetAttributesServlet gas = new GetAttributesServlet(Collections.singleton(KEY));
      sas = new SetAttributesServlet(Collections.singletonMap(KEY, getAttributeValue(attr++)));   
      MultipleActionServlet mas = new MultipleActionServlet(gas, sas);
      ReplicationListener replListener0 = getReplicationListener(mgr0.getCache());
      replListener0.expectWithTx(PutDataMapCommand.class); 
      
      req = new Request(mgr3, sessionId, mas);
      req.execute();      
      replListener0.waitForReplicationToOccur();
      
      assert sessionId.equals(mas.getSessionId()) : "wrong session id; expected " + sessionId + " got " + mas.getSessionId();
      Integer integer = (Integer) gas.getReadAttributes().get(KEY);
      assert integer != null : "null attribute value";
      assert integer.intValue() == (attr - 2) : "wrong val " + integer + " expected " + (attr -2);
      
      BuddyReplicationAssertions.assertBuddyBackup(contextHostName, sessionId, mgr3.getCache(), mgr0.getCache());
      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr1.getCache());
      
      // Invalidate the session
      InvalidationServlet invs = new InvalidationServlet();
      replListener0.expectWithTx(RemoveNodeCommand.class); 
      
      req = new Request(mgr3, sessionId, invs);
      req.execute();      
      replListener0.waitForReplicationToOccur();

      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr0.getCache());
      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr1.getCache());
      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr2.getCache());
      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr3.getCache());
   }
   
   public void testFailoverToOwnerAndImmediatelyInvalidate()
   {
      int attr = 0;
      
      SessionManager mgr0 = getSessionManagers().get(0);
      SessionManager mgr1 = getSessionManagers().get(1);
      SessionManager mgr2 = getSessionManagers().get(2);
      SessionManager mgr3 = getSessionManagers().get(3);
      
      String contextHostName = mgr0.getContextHostName();

      // Create the session
      SetAttributesServlet sas = new SetAttributesServlet(Collections.singletonMap(KEY, getAttributeValue(attr++)));
      ReplicationListener replListener2 = getReplicationListener(mgr2.getCache());
      replListener2.expectWithTx(PutDataMapCommand.class);
      
      Request req = new Request(mgr1, null, sas);
      req.execute();      
      replListener2.waitForReplicationToOccur();      
      
      String sessionId = sas.getSessionId();
      assert sessionId != null : "session id is null";
      // validate cache contents
      BuddyReplicationAssertions.assertBuddyBackup(contextHostName, sessionId, mgr1.getCache(), mgr2.getCache());
      
      // Modify the session
      sas = new SetAttributesServlet(Collections.singletonMap(KEY, getAttributeValue(attr++))); 
      replListener2.expectWithTx(PutDataMapCommand.class);
      
      req = new Request(mgr1, sessionId, sas);
      req.execute();      
      replListener2.waitForReplicationToOccur();
      
      // Fail over; request reads the session and then modifies the session
      GetAttributesServlet gas = new GetAttributesServlet(Collections.singleton(KEY));
      sas = new SetAttributesServlet(Collections.singletonMap(KEY, getAttributeValue(attr++)));   
      MultipleActionServlet mas = new MultipleActionServlet(gas, sas);
      ReplicationListener replListener1 = getReplicationListener(mgr1.getCache());
      replListener1.expectWithTx(PutDataMapCommand.class); 
      
      req = new Request(mgr0, sessionId, mas);
      req.execute();      
      replListener1.waitForReplicationToOccur();
      
      assert sessionId.equals(mas.getSessionId()) : "wrong session id; expected " + sessionId + " got " + mas.getSessionId();
      Integer integer = (Integer) gas.getReadAttributes().get(KEY);
      assert integer != null : "null attribute value";
      assert integer.intValue() == (attr - 2) : "wrong val " + integer + " expected " + (attr -2);
      
      BuddyReplicationAssertions.assertBuddyBackup(contextHostName, sessionId, mgr0.getCache(), mgr1.getCache());
      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr2.getCache());
      
      // Invalidate the session
      InvalidationServlet invs = new InvalidationServlet();
      replListener1.expectWithTx(RemoveNodeCommand.class); 
      
      req = new Request(mgr0, sessionId, invs);
      req.execute();      
      replListener1.waitForReplicationToOccur();

      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr0.getCache());
      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr1.getCache());
      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr2.getCache());
      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr3.getCache());
   }
   
   public void testFailoverAndInvalidate()
   {
      int attr = 0;
      
      SessionManager mgr0 = getSessionManagers().get(0);
      SessionManager mgr1 = getSessionManagers().get(1);
      SessionManager mgr2 = getSessionManagers().get(2);
      SessionManager mgr3 = getSessionManagers().get(3);
      
      String contextHostName = mgr0.getContextHostName();

      // Create the session
      SetAttributesServlet sas = new SetAttributesServlet(Collections.singletonMap(KEY, getAttributeValue(attr++)));
      ReplicationListener replListener1 = getReplicationListener(mgr1.getCache());
      replListener1.expectWithTx(PutDataMapCommand.class);
      
      Request req = new Request(mgr0, null, sas);
      req.execute();      
      replListener1.waitForReplicationToOccur();      
      
      String sessionId = sas.getSessionId();
      assert sessionId != null : "session id is null";
      // validate cache contents
      BuddyReplicationAssertions.assertBuddyBackup(contextHostName, sessionId, mgr0.getCache(), mgr1.getCache());
      
      // Modify the session
      sas = new SetAttributesServlet(Collections.singletonMap(KEY, getAttributeValue(attr++))); 
      replListener1.expectWithTx(PutDataMapCommand.class);
      
      req = new Request(mgr0, sessionId, sas);
      req.execute();      
      replListener1.waitForReplicationToOccur();
      
      // Fail over; request reads the session and then modifies the session
      GetAttributesServlet gas = new GetAttributesServlet(Collections.singleton(KEY));
      sas = new SetAttributesServlet(Collections.singletonMap(KEY, getAttributeValue(attr++)));   
      MultipleActionServlet mas = new MultipleActionServlet(gas, sas);
      ReplicationListener replListener0 = getReplicationListener(mgr0.getCache());
      replListener0.expectWithTx(PutDataMapCommand.class); 
      
      req = new Request(mgr3, sessionId, mas);
      req.execute();      
      replListener0.waitForReplicationToOccur();
      
      assert sessionId.equals(mas.getSessionId()) : "wrong session id; expected " + sessionId + " got " + mas.getSessionId();
      Integer integer = (Integer) gas.getReadAttributes().get(KEY);
      assert integer != null : "null attribute value";
      assert integer.intValue() == (attr - 2) : "wrong val " + integer + " expected " + (attr -2);
      
      BuddyReplicationAssertions.assertBuddyBackup(contextHostName, sessionId, mgr3.getCache(), mgr0.getCache());
      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr1.getCache());
      
      // Modify the session again
      sas = new SetAttributesServlet(Collections.singletonMap(KEY, getAttributeValue(attr++)));
      replListener0.expectWithTx(PutDataMapCommand.class);
      
      req = new Request(mgr3, sessionId, sas);
      req.execute();      
      replListener0.waitForReplicationToOccur();
      
      // Invalidate the session
      InvalidationServlet invs = new InvalidationServlet();
      replListener0.expectWithTx(RemoveNodeCommand.class); 
      
      req = new Request(mgr3, sessionId, invs);
      req.execute();      
      replListener0.waitForReplicationToOccur();

      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr0.getCache());
      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr1.getCache());
      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr2.getCache());
      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr3.getCache());
   }
   
   public void testFailoverToOwnerAndInvalidate()
   {
      int attr = 0;
      
      SessionManager mgr0 = getSessionManagers().get(0);
      SessionManager mgr1 = getSessionManagers().get(1);
      SessionManager mgr2 = getSessionManagers().get(2);
      SessionManager mgr3 = getSessionManagers().get(3);
      
      String contextHostName = mgr0.getContextHostName();

      // Create the session
      SetAttributesServlet sas = new SetAttributesServlet(Collections.singletonMap(KEY, getAttributeValue(attr++)));
      ReplicationListener replListener2 = getReplicationListener(mgr2.getCache());
      replListener2.expectWithTx(PutDataMapCommand.class);
      
      Request req = new Request(mgr1, null, sas);
      req.execute();      
      replListener2.waitForReplicationToOccur();      
      
      String sessionId = sas.getSessionId();
      assert sessionId != null : "session id is null";
      // validate cache contents
      BuddyReplicationAssertions.assertBuddyBackup(contextHostName, sessionId, mgr1.getCache(), mgr2.getCache());
      
      // Modify the session
      sas = new SetAttributesServlet(Collections.singletonMap(KEY, getAttributeValue(attr++))); 
      replListener2.expectWithTx(PutDataMapCommand.class);
      
      req = new Request(mgr1, sessionId, sas);
      req.execute();      
      replListener2.waitForReplicationToOccur();
      
      // Fail over; request reads the session and then modifies the session
      GetAttributesServlet gas = new GetAttributesServlet(Collections.singleton(KEY));
      sas = new SetAttributesServlet(Collections.singletonMap(KEY, getAttributeValue(attr++)));   
      MultipleActionServlet mas = new MultipleActionServlet(gas, sas);
      ReplicationListener replListener1 = getReplicationListener(mgr1.getCache());
      replListener1.expectWithTx(PutDataMapCommand.class); 
      
      req = new Request(mgr0, sessionId, mas);
      req.execute();      
      replListener1.waitForReplicationToOccur();
      
      assert sessionId.equals(mas.getSessionId()) : "wrong session id; expected " + sessionId + " got " + mas.getSessionId();
      Integer integer = (Integer) gas.getReadAttributes().get(KEY);
      assert integer != null : "null attribute value";
      assert integer.intValue() == (attr - 2) : "wrong val " + integer + " expected " + (attr -2);
      
      BuddyReplicationAssertions.assertBuddyBackup(contextHostName, sessionId, mgr0.getCache(), mgr1.getCache());
      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr2.getCache());
      
      // Modify the session again
      sas = new SetAttributesServlet(Collections.singletonMap(KEY, getAttributeValue(attr++)));
      replListener1.expectWithTx(PutDataMapCommand.class);
      
      req = new Request(mgr0, sessionId, sas);
      req.execute();      
      replListener1.waitForReplicationToOccur();
      
      // Invalidate the session
      InvalidationServlet invs = new InvalidationServlet();
      replListener1.expectWithTx(RemoveNodeCommand.class); 
      
      req = new Request(mgr0, sessionId, invs);
      req.execute();      
      replListener1.waitForReplicationToOccur();

      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr0.getCache());
      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr1.getCache());
      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr2.getCache());
      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr3.getCache());
   }
   
   public void testFailoverAndFailBack()
   {
      int attr = 0;
      
      SessionManager mgr0 = getSessionManagers().get(0);
      SessionManager mgr1 = getSessionManagers().get(1);
      SessionManager mgr2 = getSessionManagers().get(2);
      SessionManager mgr3 = getSessionManagers().get(3);
      
      String contextHostName = mgr0.getContextHostName();

      // Create the session
      SetAttributesServlet sas = new SetAttributesServlet(Collections.singletonMap(KEY, getAttributeValue(attr++)));
      ReplicationListener replListener1 = getReplicationListener(mgr1.getCache());
      replListener1.expectWithTx(PutDataMapCommand.class);
      
      Request req = new Request(mgr0, null, sas);
      req.execute();      
      replListener1.waitForReplicationToOccur();      
      
      String sessionId = sas.getSessionId();
      assert sessionId != null : "session id is null";
      // validate cache contents
      BuddyReplicationAssertions.assertBuddyBackup(contextHostName, sessionId, mgr0.getCache(), mgr1.getCache());
      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr2.getCache());
      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr3.getCache());
      
      // Modify the session
      sas = new SetAttributesServlet(Collections.singletonMap(KEY, getAttributeValue(attr++))); 
      replListener1.expectWithTx(PutDataMapCommand.class);
      
      req = new Request(mgr0, sessionId, sas);
      req.execute();      
      replListener1.waitForReplicationToOccur();
      
      // Fail over; request reads the session and then modifies the session
      GetAttributesServlet gas = new GetAttributesServlet(Collections.singleton(KEY));
      sas = new SetAttributesServlet(Collections.singletonMap(KEY, getAttributeValue(attr++)));   
      MultipleActionServlet mas = new MultipleActionServlet(gas, sas);
      ReplicationListener replListener0 = getReplicationListener(mgr0.getCache());
      replListener0.expectWithTx(PutDataMapCommand.class); 
      
      req = new Request(mgr3, sessionId, mas);
      req.execute();      
      replListener0.waitForReplicationToOccur();
      
      assert sessionId.equals(mas.getSessionId()) : "wrong session id; expected " + sessionId + " got " + mas.getSessionId();
      Integer integer = (Integer) gas.getReadAttributes().get(KEY);
      assert integer != null : "null attribute value";
      assert integer.intValue() == (attr - 2) : "wrong val " + integer + " expected " + (attr -2);
      
      BuddyReplicationAssertions.assertBuddyBackup(contextHostName, sessionId, mgr3.getCache(), mgr0.getCache());
      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr1.getCache());
      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr2.getCache());
      
      // Modify the session again
      sas = new SetAttributesServlet(Collections.singletonMap(KEY, getAttributeValue(attr++)));
      replListener0.expectWithTx(PutDataMapCommand.class);
      
      req = new Request(mgr3, sessionId, sas);
      req.execute();      
      replListener0.waitForReplicationToOccur();
      
      // Fail back; request reads the session and then modifies the session
      gas = new GetAttributesServlet(Collections.singleton(KEY));
      sas = new SetAttributesServlet(Collections.singletonMap(KEY, getAttributeValue(attr++)));   
      mas = new MultipleActionServlet(gas, sas);
      replListener1.expectWithTx(PutDataMapCommand.class); 
      
      req = new Request(mgr0, sessionId, mas);
      req.execute();      
      replListener1.waitForReplicationToOccur();
      
      assert sessionId.equals(mas.getSessionId()) : "wrong session id; expected " + sessionId + " got " + mas.getSessionId();
      integer = (Integer) gas.getReadAttributes().get(KEY);
      assert integer != null : "null attribute value";
      assert integer.intValue() == attr - 2 : "wrong val " + integer + " expected " + (attr -2);
      
      BuddyReplicationAssertions.assertBuddyBackup(contextHostName, sessionId, mgr0.getCache(), mgr1.getCache());
      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr3.getCache());
      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr2.getCache());
      
      // Invalidate the session
      InvalidationServlet invs = new InvalidationServlet();
      replListener1.expectWithTx(RemoveNodeCommand.class); 
      
      req = new Request(mgr0, sessionId, invs);
      req.execute();      
      replListener1.waitForReplicationToOccur();

      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr0.getCache());
      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr1.getCache());
      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr2.getCache());
      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr3.getCache());
      
      gas = new GetAttributesServlet(Collections.singleton(KEY));
      req = new Request(mgr0, sessionId, gas);
      assert gas.getReadAttributes().get(KEY) == null : "session not cleaned up";
   }
   
   public void testInvalidateOnFailoverToBackup()
   {
      int attr = 0;
      
      SessionManager mgr0 = getSessionManagers().get(0);
      SessionManager mgr1 = getSessionManagers().get(1);
      SessionManager mgr2 = getSessionManagers().get(2);
      SessionManager mgr3 = getSessionManagers().get(3);
      
      String contextHostName = mgr0.getContextHostName();

      // Create the session
      SetAttributesServlet sas = new SetAttributesServlet(Collections.singletonMap(KEY, getAttributeValue(attr++)));
      ReplicationListener replListener0 = getReplicationListener(mgr0.getCache());
      replListener0.expectWithTx(PutDataMapCommand.class);
      
      Request req = new Request(mgr3, null, sas);
      req.execute();      
      replListener0.waitForReplicationToOccur();      
      
      String sessionId = sas.getSessionId();
      assert sessionId != null : "session id is null";
      // validate cache contents
      BuddyReplicationAssertions.assertBuddyBackup(contextHostName, sessionId, mgr3.getCache(), mgr0.getCache());
      
      // Modify the session
      sas = new SetAttributesServlet(Collections.singletonMap(KEY, getAttributeValue(attr++))); 
      replListener0.expectWithTx(PutDataMapCommand.class);
      
      req = new Request(mgr3, sessionId, sas);
      req.execute();      
      replListener0.waitForReplicationToOccur();      
      
      // Passivate the session
      mgr0.passivate(sessionId);
      mgr1.passivate(sessionId);
      mgr2.passivate(sessionId);
      mgr3.passivate(sessionId);   
      
      // Reactivate the session
      sas = new SetAttributesServlet(Collections.singletonMap(KEY, getAttributeValue(attr++))); 
      replListener0.expectWithTx(PutDataMapCommand.class);
      
      req = new Request(mgr3, sessionId, sas);
      req.execute();      
      replListener0.waitForReplicationToOccur();      
      
      // Invalidate the session
      sas = new SetAttributesServlet(Collections.singletonMap(KEY, getAttributeValue(attr++)));
      InvalidationServlet invs = new InvalidationServlet();
      MultipleActionServlet mas = new MultipleActionServlet(sas, invs);
      ReplicationListener replListener1 = getReplicationListener(mgr1.getCache());
      replListener1.expectWithTx(PutDataMapCommand.class, RemoveNodeCommand.class); 
      
      req = new Request(mgr0, sessionId, mas);
      req.execute();      
      replListener1.waitForReplicationToOccur();

      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr0.getCache());
      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr1.getCache());
      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr2.getCache());
      BuddyReplicationAssertions.assertUnrelated(contextHostName, sessionId, mgr3.getCache());
      
      GetAttributesServlet gas = new GetAttributesServlet(Collections.singleton(KEY));
      req = new Request(mgr0, sessionId, gas);
      assert gas.getReadAttributes().get(KEY) == null : "session not cleaned up";
   }
}
